@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2016	PgPool Global Development Group
+ * Copyright (c) 2003-2012	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -35,6 +35,7 @@
 #endif
 
 #include <sys/stat.h>
+#include <sys/types.h>
 #include <fcntl.h>
 
 #include <sys/wait.h>
@@ -59,7 +60,8 @@
 #include "parser/pool_memory.h"
 #include "parser/pool_string.h"
 #include "pool_passwd.h"
-
+#include "pool_memqcache.h"
+#include "watchdog/wd_ext.h"
 /*
  * Process pending signal actions.
  */
@@ -106,7 +108,7 @@ static void reload_config(void);
 static int pool_pause(struct timeval *timeout);
 static void kill_all_children(int sig);
 static int get_next_master_node(void);
-static pid_t fork_follow_child(int old_master, int new_primary, int old_primary);
+static pid_t fork_follow_child(int old_master, int new_master, int old_primary);
 
 static RETSIGTYPE exit_handler(int sig);
 static RETSIGTYPE reap_handler(int sig);
@@ -115,6 +117,9 @@ static RETSIGTYPE reload_config_handler(int sig);
 static RETSIGTYPE health_check_timer_handler(int sig);
 static RETSIGTYPE wakeup_handler(int sig);
 
+static int health_check(void);
+static int system_db_health_check(void);
+
 static void usage(void);
 static void show_version(void);
 static void stop_me(void);
@@ -129,7 +134,6 @@ static struct sockaddr_un un_addr;		/* unix domain socket path */
 static struct sockaddr_un pcp_un_addr;  /* unix domain socket path for PCP */
 
 ProcessInfo *process_info;	/* Per child info table on shmem */
-struct timeval random_start_time;
 
 /*
  * Private copy of backend status
@@ -162,8 +166,6 @@ static int not_detach = 0;		/* non 0 if non detach option (-n) is given */
 
 static int stop_sig = SIGTERM;	/* stopping signal default value */
 
-static volatile sig_atomic_t health_check_timer_expired;		/* non 0 if health check timer expired */
-
 POOL_REQUEST_INFO *Req_info;		/* request info area in shared memory */
 volatile sig_atomic_t *InRecovery; /* non 0 if recovery is started */
 volatile sig_atomic_t reload_config_request = 0;
@@ -178,6 +180,7 @@ int my_proc_id;
 static BackendStatusRecord backend_rec;	/* Backend status record */
 
 static pid_t worker_pid; /* pid of worker process */
+static pid_t watchdog_pid; /* pid of watchdog process */
 
 BACKEND_STATUS* my_backend_status[MAX_NUM_BACKENDS];		/* Backend status buffer */
 int my_master_node_id;		/* Master node id buffer */
@@ -199,6 +202,8 @@ int main(int argc, char **argv)
 	int debug_level = 0;
 	int	optindex;
 	bool discard_status = false;
+	bool retrying;
+	bool clear_memcache_oidmaps = false;
 
 	static struct option long_options[] = {
 		{"hba-file", required_argument, NULL, 'a'},
@@ -210,6 +215,7 @@ int main(int argc, char **argv)
 		{"mode", required_argument, NULL, 'm'},
 		{"dont-detach", no_argument, NULL, 'n'},
 		{"discard-status", no_argument, NULL, 'D'},
+		{"clear-oidmaps", no_argument, NULL, 'C'},
 		{"version", no_argument, NULL, 'v'},
 		{NULL, 0, NULL, 0}
 	};
@@ -221,7 +227,7 @@ int main(int argc, char **argv)
 	snprintf(pcp_conf_file, sizeof(pcp_conf_file), "%s/%s", DEFAULT_CONFIGDIR, PCP_PASSWD_FILE_NAME);
 	snprintf(hba_file, sizeof(hba_file), "%s/%s", DEFAULT_CONFIGDIR, HBA_CONF_FILE_NAME);
 
-    while ((opt = getopt_long(argc, argv, "a:cdf:F:hm:nDv", long_options, &optindex)) != -1)
+    while ((opt = getopt_long(argc, argv, "a:cdf:F:hm:nDCv", long_options, &optindex)) != -1)
 	{
 		switch (opt)
 		{
@@ -292,6 +298,10 @@ int main(int argc, char **argv)
 				discard_status = true;
 				break;
 
+			case 'C': /* discard caches in memcached */
+				clear_memcache_oidmaps = true;
+				break;
+
 			case 'v':
 				show_version();
 				exit(0);
@@ -302,9 +312,6 @@ int main(int argc, char **argv)
 		}
 	}
 
-	/* For PostmasterRandom */
-	gettimeofday(&random_start_time, NULL);
-
 #ifdef USE_SSL
 	/* global ssl init */
 	SSL_library_init();
@@ -317,7 +324,7 @@ int main(int argc, char **argv)
 		exit(1);
 
 	/*
-	 * Init debug level with -d option value
+	 * Override debug level
 	 */
 	pool_config->debug_level = debug_level;
 
@@ -337,6 +344,12 @@ int main(int argc, char **argv)
 		pool_config->logsyslog = 1;
 	}
 
+	/*
+	 * Override debug level
+	 */
+	if (pool_config->debug_level == 0)
+		pool_config->debug_level = debug_level;
+
 	if (pool_config->enable_pool_hba)
 		load_hba(hba_file);
 
@@ -369,6 +382,7 @@ int main(int argc, char **argv)
 		if (!strcmp(argv[optind], "stop"))
 		{
 			stop_me();
+			unlink(pool_config->pid_file_name);
 			pool_shmem_exit(0);
 			exit(0);
 		}
@@ -405,6 +419,28 @@ int main(int argc, char **argv)
 		exit(1);
 	}
 
+	/* check effective user id for watchdog */
+	/* watchdog must be started under the privileged user */
+	if (pool_config->use_watchdog )
+	{
+		/* check sticky bit of network interface control commands */
+		if (wd_chk_sticky() == 1)
+		{
+			/* if_up, if_down and arping command have a sticky bit */
+			pool_log("watchdog might call network commands which using sticky bit.");
+		}
+		else
+		{
+			/* check root */
+			if (geteuid() != 0)
+			{
+				pool_error("watchdog must be started under the privileged user ID to up/down virtual network interface.");
+				pool_shmem_exit(1);
+				exit(1);
+			}
+		}
+	}
+
 	/* set signal masks */
 	poolinitmask();
 
@@ -415,6 +451,8 @@ int main(int argc, char **argv)
 
 	/*
 	 * Locate pool_passwd
+	 * The default file name "pool_passwd" can be changed by setting
+	 * pgpool.conf's "pool_passwd" directive.
 	 */
 	if (strcmp("", pool_config->pool_passwd))
 	{
@@ -532,7 +570,78 @@ int main(int argc, char **argv)
 		pool_error("failed to allocate InRecovery");
 		myexit(1);
 	}
-	*InRecovery = 0;
+	*InRecovery = RECOVERY_INIT;
+
+	/*
+	 * Initialize shared memory cache
+	 */
+	if (pool_config->memory_cache_enabled)
+	{
+		if (pool_is_shmem_cache())
+		{
+			size_t size;
+			
+			size = pool_shared_memory_cache_size();
+			if (size == 0)
+			{
+				pool_error("pool_shared_memory_cache_size error");
+				myexit(1);
+			}
+
+			if (pool_init_memory_cache(size) < 0)
+			{
+				pool_error("pool_shared_memory_cache_size error");
+				myexit(1);
+			}
+
+			size = pool_shared_memory_fsmm_size();
+			if (size == 0)
+			{
+				pool_error("pool_shared_memory_fsmm_size error");
+				myexit(1);
+			}
+			pool_init_fsmm(size);
+
+			pool_allocate_fsmm_clock_hand();
+
+			pool_discard_oid_maps();
+			pool_log("pool_discard_oid_maps: discarded memqcache oid maps");
+
+			pool_hash_init(pool_config->memqcache_max_num_cache);
+		}
+
+#ifdef USE_MEMCACHED
+		else
+		{
+			if (clear_memcache_oidmaps)
+			{
+				pool_discard_oid_maps();
+				pool_log("pool_discard_oid_maps: discarded memqcache oid maps");
+			}
+			else
+			{
+				pool_debug("skipped discarding memqcache oid maps");
+			}
+		}
+#endif
+
+		if (pool_init_memqcache_stats() < 0)
+		{
+			pool_error("pool_init_memqcache_stats error");
+			myexit(1);
+		}
+	}
+
+	/* start watchdog */
+	if (pool_config->use_watchdog )
+	{
+		watchdog_pid = wd_main(1);
+		if (watchdog_pid == 0)
+		{
+			pool_error("wd_main error");
+			myexit(1);
+		}
+	}
 
 	/*
 	 * We need to block signal here. Otherwise child might send some
@@ -589,6 +698,9 @@ int main(int argc, char **argv)
 	/*
 	 * This is the main loop
 	 */
+
+	retrying = false;
+
 	for (;;)
 	{
 		CHECK_REQUEST;
@@ -628,6 +740,7 @@ int main(int argc, char **argv)
 			POOL_SETMASK(&UnBlockSig);
 			sts = health_check();
 			POOL_SETMASK(&BlockSig);
+
 			if (pool_config->parallel_mode || pool_config->enable_query_cache)
 				sys_sts = system_db_health_check();
 
@@ -637,34 +750,48 @@ int main(int argc, char **argv)
 				{
 					sts--;
 
+					retrycnt++;
+					pool_signal(SIGALRM, SIG_IGN);	/* Cancel timer */
+
 					if (!pool_config->parallel_mode)
 					{
 						if (POOL_DISALLOW_TO_FAILOVER(BACKEND_INFO(sts).flag))
 						{
 							pool_log("health_check: %d failover is canceld because failover is disallowed", sts);
 						}
+						else if (retrycnt <= pool_config->health_check_max_retries)
+						{
+							/* continue to retry */
+							sleep_time = pool_config->health_check_retry_delay;
+							pool_log("health check retry sleep time: %d second(s)", sleep_time);
+							pool_sleep(sleep_time);
+							retrying = true;
+							continue;
+						}
 						else
 						{
+							/* retry count over */
 							pool_log("set %d th backend down status", sts);
 							Req_info->kind = NODE_DOWN_REQUEST;
 							Req_info->node_id[0] = sts;
+							health_check_timer_expired = 0;
 							failover();
 							/* need to distribute this info to children */
+							retrying = false;
 						}
 					}
 					else
 					{
-						retrycnt++;
-						pool_signal(SIGALRM, SIG_IGN);	/* Cancel timer */
-
 						if (retrycnt > NUM_BACKENDS)
 						{
 							/* retry count over */
 							pool_log("set %d th backend down status", sts);
 							Req_info->kind = NODE_DOWN_REQUEST;
 							Req_info->node_id[0] = sts;
+							health_check_timer_expired = 0;
 							failover();
 							retrycnt = 0;
+							retrying = false;
 						}
 						else
 						{
@@ -696,6 +823,16 @@ int main(int argc, char **argv)
 					}
 				}
 			}
+			else
+			{
+				/* success. reset retry count */
+				retrycnt = 0;
+				if (retrying)
+				{
+					pool_log("after some retrying backend returned to healthy state");
+					retrying = false;
+				}
+			}
 
 			if (pool_config->health_check_timeout > 0)
 			{
@@ -753,6 +890,8 @@ static void usage(void)
 	fprintf(stderr, "  -h, --help          Prints this help\n\n");
 	fprintf(stderr, "Start options:\n");
 	fprintf(stderr, "  -c, --clear         Clears query cache (enable_query_cache must be on)\n");
+	fprintf(stderr, "  -C, --clear-oidmaps Clears query cache oidmaps when memqcache_method is memcached\n");
+	fprintf(stderr, "                      (If shmem, discards whenever pgpool starts.)\n");
 	fprintf(stderr, "  -n, --dont-detach   Don't run in daemon mode, does not detach control tty\n");
 	fprintf(stderr, "  -D, --discard-status Discard pgpool_status file and do not restore previous status\n");
 	fprintf(stderr, "  -d, --debug         Debug mode\n\n");
@@ -799,21 +938,14 @@ static void daemonize(void)
 #endif
 
 	mypid = getpid();
-
+	write_pid_file();
 	rc_chdir = chdir("/");
 
 	i = open("/dev/null", O_RDWR);
-	if(i < 0)
-	{
-		pool_error("Failed to open \"/dev/null\" reason:%s", strerror(errno));
-	}
-	else
-	{
-		dup2(i, 0);
-		dup2(i, 1);
-		dup2(i, 2);
-		close(i);
-	}
+	dup2(i, 0);
+	dup2(i, 1);
+	dup2(i, 2);
+
 	/* close syslog connection for daemonizing */
 	if (pool_config->logsyslog) {
 		closelog();
@@ -828,7 +960,6 @@ static void daemonize(void)
 		openlog(pool_config->syslog_ident, LOG_PID|LOG_NDELAY|LOG_NOWAIT, pool_config->syslog_facility);
 	}
 
-	write_pid_file();
 }
 
 
@@ -1017,7 +1148,6 @@ static int read_status_file(bool discard_status)
 static int write_status_file(void)
 {
 	FILE *fd;
-	int fdnum;
 	char fnamebuf[POOLMAXPATHLEN];
 	int i;
 
@@ -1043,32 +1173,6 @@ static int write_status_file(void)
 		fclose(fd);
 		return -1;
 	}
-
-	if (fflush(fd) != 0)
-	{
-		pool_error("Could not flush backend status file as %s. reason: %s",
-				   fnamebuf, strerror(errno));
-		fclose(fd);
-		return -1;
-	}
-
-	fdnum = fileno(fd);
-	if (fdnum < 0)
-	{
-		pool_error("Failed to get file number of %s. reason: %s",
-				   fnamebuf, strerror(errno));
-		fclose(fd);
-		return -1;
-	}
-
-	if (fsync(fdnum) != 0)
-	{
-		pool_error("Failed to fsync() %s. reason: %s",
-				   fnamebuf, strerror(errno));
-		fclose(fd);
-		return -1;
-	}
-
 	fclose(fd);
 	return 0;
 }
@@ -1309,7 +1413,6 @@ static void myunlink(const char* path)
 static void myexit(int code)
 {
 	int i;
-	pid_t wpid;
 
 	if (getpid() != mypid)
 		return;
@@ -1325,18 +1428,10 @@ static void myexit(int code)
 				kill(pid, SIGTERM);
 			}
 		}
-
-		/* wait for all children to exit */
-		do
-		{
-			pid_t ret_pid;
-
-			wpid = waitpid(-1, &ret_pid, WNOHANG);
-		} while (wpid > 0 || (wpid == -1 && errno == EINTR));
-
-		if (wpid == -1 && errno != ECHILD)
+		while (wait(NULL) > 0)
+			;
+		if (errno != ECHILD)
 			pool_error("wait() failed. reason:%s", strerror(errno));
-
 		POOL_SETMASK(&UnBlockSig);
 	}
 
@@ -1357,21 +1452,8 @@ void notice_backend_error(int node_id)
 	degenerate_backend_set(&n, 1);
 }
 
-/*
- * degenerate_backend_set_ex:
- *
- * The function signals/verifies the node down request.
- * The request is then processed by failover function.
- *
- * node_id_set: array of node ids to be registered for NODE DOWN operation
- * count:       number of elements in node_id_set array
- * test_only:   When set, function only checks if NODE DOWN operation can be
- *              executed on provided node ids and never registers the operation
- *              request.
- *              For test_only case function returs false as
- *              soon as first non complient node in node_id_set is found
- */
-bool degenerate_backend_set_ex(int *node_id_set, int count, bool test_only)
+/* notice backend connection error using SIGUSR1 */
+void degenerate_backend_set(int *node_id_set, int count)
 {
 	pid_t parent = getppid();
 	int i;
@@ -1384,64 +1466,42 @@ bool degenerate_backend_set_ex(int *node_id_set, int count, bool test_only)
 
 	if (pool_config->parallel_mode)
 	{
-		return false;
-	}
-	if (test_only == false)
-	{
-		POOL_SETMASK2(&BlockSig, &oldmask);
-		pool_semaphore_lock(REQUEST_INFO_SEM);
-		Req_info->kind = NODE_DOWN_REQUEST;
+		return;
 	}
 
+	POOL_SETMASK2(&BlockSig, &oldmask);
+	pool_semaphore_lock(REQUEST_INFO_SEM);
+	Req_info->kind = NODE_DOWN_REQUEST;
 	for (i = 0; i < count; i++)
 	{
 		if (node_id_set[i] < 0 || node_id_set[i] >= MAX_NUM_BACKENDS ||
 			!VALID_BACKEND(node_id_set[i]))
 		{
 			pool_log("degenerate_backend_set: node %d is not valid backend.", i);
-			if (test_only)
-				return false;
 			continue;
 		}
 
 		if (POOL_DISALLOW_TO_FAILOVER(BACKEND_INFO(node_id_set[i]).flag))
 		{
 			pool_log("degenerate_backend_set: %d failover request from pid %d is canceld because failover is disallowed", node_id_set[i], getpid());
-			if (test_only)
-				return false;
 			continue;
 		}
 
-		if (test_only == false)
-		{
-			pool_log("degenerate_backend_set: %d fail over request from pid %d", node_id_set[i], getpid());
-			Req_info->node_id[i] = node_id_set[i];
-		}
+		pool_log("degenerate_backend_set: %d fail over request from pid %d", node_id_set[i], getpid());
+		Req_info->node_id[i] = node_id_set[i];
 		need_signal = true;
 	}
 
 	if (need_signal)
 	{
-		if (test_only)
-			return true;
-		kill(parent, SIGUSR1);
+		if (!pool_config->use_watchdog || WD_OK == wd_degenerate_backend_set(node_id_set, count))
+		{
+			kill(parent, SIGUSR1);
+		}
 	}
-	else if (test_only)
-		return false;
 
 	pool_semaphore_unlock(REQUEST_INFO_SEM);
 	POOL_SETMASK(&oldmask);
-
-	return need_signal;
-}
-
-/*
- * wrapper over degenerate_backend_set_ex function to signal
- * NODE down operation request
- */
-void degenerate_backend_set(int *node_id_set, int count)
-{
-	degenerate_backend_set_ex(node_id_set, count, false);
 }
 
 /* send promote node request using SIGUSR1 */
@@ -1464,7 +1524,11 @@ void promote_backend(int node_id)
 	Req_info->kind = PROMOTE_NODE_REQUEST;
 	Req_info->node_id[0] = node_id;
 	pool_log("promote_backend: %d promote node request from pid %d", node_id, getpid());
-	kill(parent, SIGUSR1);
+
+	if (!pool_config->use_watchdog || WD_OK == wd_promote_backend(node_id))
+	{
+		kill(parent, SIGUSR1);
+	}
 	pool_semaphore_unlock(REQUEST_INFO_SEM);
 }
 
@@ -1484,13 +1548,16 @@ void send_failback_request(int node_id)
 		return;
 	}
 
+	if (pool_config->use_watchdog && WD_OK != wd_send_failback_request(node_id))
+	{
+		return;
+	}
 	kill(parent, SIGUSR1);
 }
 
 static RETSIGTYPE exit_handler(int sig)
 {
 	int i;
-	pid_t wpid;
 
 	POOL_SETMASK(&AuthBlockSig);
 
@@ -1502,6 +1569,7 @@ static RETSIGTYPE exit_handler(int sig)
 	{
 		pool_debug("exit_handler: I am not parent");
 		POOL_SETMASK(&UnBlockSig);
+		pool_shmem_exit(0);
 		exit(0);
 	}
 
@@ -1520,11 +1588,6 @@ static RETSIGTYPE exit_handler(int sig)
 
 	exiting = 1;
 
-	/* Close listen socket */
-	pool_log("pgpool main: close listen socket");
-	close(inet_fd);
-	close(unix_fd);
-
 	for (i = 0; i < pool_config->num_init_children; i++)
 	{
 		pid_t pid = process_info[i].pid;
@@ -1537,16 +1600,18 @@ static RETSIGTYPE exit_handler(int sig)
 	kill(pcp_pid, sig);
 	kill(worker_pid, sig);
 
+	if (pool_config->use_watchdog)
+	{
+		pool_log("watchdog_pid: %d", watchdog_pid);
+		kill(watchdog_pid, sig);
+	}
+
 	POOL_SETMASK(&UnBlockSig);
 
-    do
-    {
-		pid_t ret_pid;
+	while (wait(NULL) > 0)
+		;
 
-        wpid = waitpid(-1, &ret_pid, WNOHANG);
-    } while (wpid > 0 || (wpid == -1 && errno == EINTR));
-
-    if (wpid == -1 && errno != ECHILD)
+	if (errno != ECHILD)
 		pool_error("wait() failed. reason:%s", strerror(errno));
 
 	process_info = NULL;
@@ -1567,12 +1632,10 @@ static int get_next_master_node(void)
 		 * Do not use VALID_BACKEND macro in raw mode.
 		 * VALID_BACKEND return true only if the argument is master
 		 * node id. In other words, standby nodes are false. So need
-		 * to check backend status without VALID_BACKEND.
+		 * to check backend status with VALID_BACKEND_RAW.
 		 */
-		if (RAW_MODE)
+		if (RAW_MODE && VALID_BACKEND_RAW(i))
 		{
-			/* if in raw mode, we can switch to a standby node */
-			if (BACKEND_INFO(i).backend_status == CON_CONNECT_WAIT)
 				break;
 		}
 		else if (VALID_BACKEND(i))
@@ -1609,8 +1672,6 @@ static void failover(void)
 	int new_primary;
 	int nodes[MAX_NUM_BACKENDS];
 	bool need_to_restart_children;
-	int status;
-	int sts;
 
 	pool_debug("failover_handler called");
 
@@ -1669,18 +1730,14 @@ static void failover(void)
 	/* failback request? */
 	if (Req_info->kind == NODE_UP_REQUEST)
 	{
-		if (node_id < 0 || node_id >= MAX_NUM_BACKENDS ||
+		if (node_id >= MAX_NUM_BACKENDS ||
 			(Req_info->kind == NODE_UP_REQUEST && !(RAW_MODE &&
             BACKEND_INFO(node_id).backend_status == CON_DOWN) && VALID_BACKEND(node_id)) ||
 			(Req_info->kind == NODE_DOWN_REQUEST && !VALID_BACKEND(node_id)))
 		{
 			pool_semaphore_unlock(REQUEST_INFO_SEM);
-
-			if (node_id < 0 || node_id >= MAX_NUM_BACKENDS)
-				pool_error("failover_handler: invalid node_id %d MAX_NUM_BACKENDS: %d", node_id, MAX_NUM_BACKENDS);
-			else
-				pool_error("failover_handler: invalid node_id %d status:%d MAX_NUM_BACKENDS: %d", node_id,
-						   BACKEND_INFO(node_id).backend_status, MAX_NUM_BACKENDS);
+			pool_error("failover_handler: invalid node_id %d status:%d MAX_NUM_BACKENDS: %d", node_id,
+					   BACKEND_INFO(node_id).backend_status, MAX_NUM_BACKENDS);
 			kill(pcp_pid, SIGUSR2);
 			switching = 0;
 			Req_info->switching = false;
@@ -1719,13 +1776,13 @@ static void failover(void)
 		for (i = 0; i < MAX_NUM_BACKENDS; i++)
 		{
 			if (Req_info->node_id[i] != -1 &&
-				VALID_BACKEND(Req_info->node_id[i]))
+				((RAW_MODE && VALID_BACKEND_RAW(Req_info->node_id[i])) ||
+				 VALID_BACKEND(Req_info->node_id[i])))
 			{
 				pool_log("starting degeneration. shutdown host %s(%d)",
 						 BACKEND_INFO(Req_info->node_id[i]).backend_hostname,
 						 BACKEND_INFO(Req_info->node_id[i]).backend_port);
 
-
 				BACKEND_INFO(Req_info->node_id[i]).backend_status = CON_DOWN;	/* set down status */
 				/* save down node */
 				nodes[Req_info->node_id[i]] = 1;
@@ -1766,7 +1823,7 @@ static void failover(void)
 #ifdef NOT_USED
 	else
 	{
-		if (Req_info->master_node_id == new_master && *InRecovery == 0)
+		if (Req_info->master_node_id == new_master && *InRecovery == RECOVERY_INIT)
 		{
 			pool_log("failover_handler: do not restart pgpool. same master node %d was selected", new_master);
 			if (Req_info->kind == NODE_UP_REQUEST)
@@ -1805,7 +1862,7 @@ static void failover(void)
 	* and request is NODE_UP_REQUEST(failback case) we don't need to
 	* restart all children. Existing session will not use newly
 	* attached node, but load balanced node is not changed util this
-	* session ends, so it's not harmless anyway.
+	* session ends, so it's harmless anyway.
 	*/
 	if (MASTER_SLAVE && !strcmp(pool_config->master_slave_sub_mode, MODE_STREAMREP)	&&
 		Req_info->kind == NODE_UP_REQUEST)
@@ -1853,21 +1910,6 @@ static void failover(void)
 
 	if (Req_info->kind == PROMOTE_NODE_REQUEST && VALID_BACKEND(node_id))
 		new_primary = node_id;
-
-	/*
-	 * If the down node was a standby node in streaming replication
-	 * mode, we can avoid calling find_primary_node_repeatedly() and
-	 * recognize the former primary as the new primary node, which
-	 * will reduce the time to process standby down.
-	 */
-	else if (MASTER_SLAVE && !strcmp(pool_config->master_slave_sub_mode, MODE_STREAMREP) &&
-			 Req_info->kind == NODE_DOWN_REQUEST)
-	{
-		if (Req_info->primary_node_id != node_id)
-			new_primary = Req_info->primary_node_id;
-		else
-			new_primary =  find_primary_node_repeatedly();
-	}
 	else
 		new_primary =  find_primary_node_repeatedly();
 
@@ -1922,7 +1964,7 @@ static void failover(void)
 	/* exec follow_master_command */
 	if ((follow_cnt > 0) && (*pool_config->follow_master_command != '\0'))
 	{
-		follow_pid = fork_follow_child(Req_info->master_node_id, new_primary,
+		follow_pid = fork_follow_child(Req_info->master_node_id, new_master,
 									   Req_info->primary_node_id);
 	}
 
@@ -1995,29 +2037,6 @@ static void failover(void)
 	 * Send restart request to pcp child.
 	 */
 	kill(pcp_pid, SIGUSR1);
-	for (;;)
-	{
-		sts = waitpid(pcp_pid, &status, 0);
-		if (sts != -1)
-			break;
-		if (sts == -1)
-		{
-			if (errno == EINTR)
-				continue;
-			else
-			{
-				pool_error("failover: waitpid failed. reson: %s", strerror(errno));
-				return;
-			}
-		}
-	}
-	if (WIFSIGNALED(status))
-		pool_log("PCP child %d exits with status %d by signal %d in failover()", pcp_pid, status, WTERMSIG(status));
-	else
-		pool_log("PCP child %d exits with status %d in failover()", pcp_pid, status);
-
-	pcp_pid = pcp_fork_a_child(pcp_unix_fd, pcp_inet_fd, pcp_conf_file);
-	pool_log("fork a new PCP child pid %d in failover()", pcp_pid);
 }
 
 /*
@@ -2032,23 +2051,15 @@ static RETSIGTYPE health_check_timer_handler(int sig)
 
 
 /*
- * check if we can connect to the backend
+ * Check if we can connect to the backend
  * returns 0 for ok. otherwise returns backend id + 1
  */
-int health_check(void)
+static int health_check(void)
 {
-	int fd;
-	int sts;
+	POOL_CONNECTION_POOL_SLOT *slot;
+	BackendInfo *bkinfo;
 	static bool is_first = true;
 	static char *dbname;
-
-	/* V2 startup packet */
-	typedef struct {
-		int len;		/* startup packet length */
-		StartupPacket_v2 sp;
-	} MySp;
-	MySp mysp;
-	char kind;
 	int i;
 
 	/* Do not execute health check during recovery */
@@ -2062,79 +2073,38 @@ int health_check(void)
 	if (is_first)
 		dbname = "postgres";
 
-	memset(&mysp, 0, sizeof(mysp));
-	mysp.len = htonl(296);
-	mysp.sp.protoVersion = htonl(PROTO_MAJOR_V2 << 16);
-	strcpy(mysp.sp.database, dbname);
-	strncpy(mysp.sp.user, pool_config->health_check_user, sizeof(mysp.sp.user) - 1);
-	*mysp.sp.options = '\0';
-	*mysp.sp.unused = '\0';
-	*mysp.sp.tty = '\0';
-
 	for (i=0;i<pool_config->backend_desc->num_backends;i++)
 	{
-		pool_debug("health_check: %d th DB node status: %d", i, BACKEND_INFO(i).backend_status);
+		/*
+		 * Make sure that health check timer has not been expired.
+		 * Before called health_check(), health_check_timer_expired is
+		 * set to 0.  However it is possible that while processing DB
+		 * nodes health check timer expired.
+		 */
+		if (health_check_timer_expired)
+		{
+			pool_log("health_check: health check timer has been already expired before attempting to connect to %d th backend", i);
+			return i+1;
+		}
 
-		if (BACKEND_INFO(i).backend_status == CON_UNUSED ||
-			BACKEND_INFO(i).backend_status == CON_DOWN)
+		bkinfo = pool_get_node_info(i);
+
+		pool_debug("health_check: %d th DB node status: %d", i, bkinfo->backend_status);
+
+		if (bkinfo->backend_status == CON_UNUSED ||
+			bkinfo->backend_status == CON_DOWN)
 			continue;
 
-		if (*(BACKEND_INFO(i).backend_hostname) == '/')
-			fd = connect_unix_domain_socket(i, FALSE);
-		else
-			fd = connect_inet_domain_socket(i, FALSE);
-
-		if (fd < 0)
-		{
-			pool_error("health check failed. %d th host %s at port %d is down",
-					   i,
-					   BACKEND_INFO(i).backend_hostname,
-					   BACKEND_INFO(i).backend_port);
-
-			return i+1;
-		}
-
-		if (write(fd, &mysp, sizeof(mysp)) < 0)
-		{
-			pool_error("health check failed during write. host %s at port %d is down. reason: %s",
-					   BACKEND_INFO(i).backend_hostname,
-					   BACKEND_INFO(i).backend_port,
-					   strerror(errno));
-			close(fd);
-			return i+1;
-		}
-
-		/*
-		 * Don't bother to be blocked by read(2). It will be
-		 * interrupted by ALRAM anyway.
-		 */
-		sts = read(fd, &kind, 1);
-		if (sts == -1)
-		{
-			pool_error("health check failed during read. host %s at port %d is down. reason: %s",
-					   BACKEND_INFO(i).backend_hostname,
-					   BACKEND_INFO(i).backend_port,
-					   strerror(errno));
-			close(fd);
-			return i+1;
-		}
-		else if (sts == 0)
-		{
-			pool_error("health check failed. EOF encountered. host %s at port %d is down",
-					   BACKEND_INFO(i).backend_hostname,
-					   BACKEND_INFO(i).backend_port);
-			close(fd);
-			return i+1;
-		}
+		slot = make_persistent_db_connection(bkinfo->backend_hostname, 
+											 bkinfo->backend_port,
+											 dbname,
+											 pool_config->health_check_user,
+											 pool_config->health_check_password, false);
 
 		if (is_first)
 			is_first = false;
 
-		/*
-		 * If a backend raised a FATAL error(max connections error or
-		 * starting up error?), do not send a Terminate message.
-		 */
-		if ((kind != 'E') && (write(fd, "X", 1) < 0))
+		if (!slot)
 		{
 			if (!strcmp(dbname, "postgres"))
 			{
@@ -2144,16 +2114,19 @@ int health_check(void)
 				dbname = "template1";
 				goto Retry;
 			}
-
-			pool_error("health check failed during write. host %s at port %d is down. reason: %s. Perhaps wrong health check user?",
-					   BACKEND_INFO(i).backend_hostname,
-					   BACKEND_INFO(i).backend_port,
-					   strerror(errno));
-			close(fd);
-			return i+1;
+			else
+			{
+				pool_error("health check failed. %d th host %s at port %d is down",
+						   i,
+						   bkinfo->backend_hostname,
+						   bkinfo->backend_port);
+				return i+1;
+			}
+		}
+		else
+		{
+			discard_persistent_db_connection(slot);
 		}
-
-		close(fd);
 	}
 
 	return 0;
@@ -2163,7 +2136,7 @@ int health_check(void)
  * check if we can connect to the SystemDB
  * returns 0 for ok. otherwise returns -1
  */
-int
+static int
 system_db_health_check(void)
 {
 	int fd;
@@ -2291,6 +2264,7 @@ static void reaper(void)
 
 			pcp_pid = pcp_fork_a_child(pcp_unix_fd, pcp_inet_fd, pcp_conf_file);
 			pool_log("fork a new PCP child pid %d", pcp_pid);
+			break;
 		}
 
 		/* exiting process was worker process */
@@ -2305,8 +2279,8 @@ static void reaper(void)
 				worker_pid = worker_fork_a_child();
 
 			pool_log("fork a new worker child pid %d", worker_pid);
-		}
-		else
+			break;
+		} else
 		{
 			if (WIFSIGNALED(status))
 				pool_debug("child %d exits with status %d by signal %d", pid, status, WTERMSIG(status));
@@ -2559,7 +2533,7 @@ static int trigger_failover_command(int node, const char *command_line,
 		return 0;
 
 	/* check failed nodeID */
-	if (node < 0 || node >= NUM_BACKENDS)
+	if (node < 0 || node > NUM_BACKENDS)
 		return -1;
 
 	info = pool_get_node_info(node);
@@ -2603,7 +2577,7 @@ static int trigger_failover_command(int node, const char *command_line,
 						break;
 
 					case 'H': /* new master host name */
-						newmaster = pool_get_node_info(new_master);
+						newmaster = pool_get_node_info(get_next_master_node());
 						if (newmaster)
 							string_append_char(exec_cmd, newmaster->backend_hostname);
 						else
@@ -2616,6 +2590,27 @@ static int trigger_failover_command(int node, const char *command_line,
 						string_append_char(exec_cmd, port_buf);
 						break;
 
+					case 'r': /* new master port */
+						newmaster = pool_get_node_info(get_next_master_node());
+						if (newmaster)
+						{
+							snprintf(port_buf, sizeof(port_buf), "%d", newmaster->backend_port);
+							string_append_char(exec_cmd, port_buf);
+						}
+						else
+							/* no vaid new master */
+							string_append_char(exec_cmd, "");
+						break;
+
+					case 'R': /* new master database directory */
+						newmaster = pool_get_node_info(get_next_master_node());
+						if (newmaster)
+							string_append_char(exec_cmd, newmaster->backend_data_directory);
+						else
+							/* no vaid new master */
+							string_append_char(exec_cmd, "");
+						break;
+
 					case 'M': /* old master node id */
 						snprintf(port_buf, sizeof(port_buf), "%d", old_master);
 						string_append_char(exec_cmd, port_buf);
@@ -2663,6 +2658,7 @@ static int find_primary_node(void)
 	BackendInfo *bkinfo;
 	POOL_CONNECTION_POOL_SLOT *s;
 	POOL_CONNECTION *con; 
+	POOL_STATUS status;
 	POOL_SELECT_RESULT *res;
 	bool is_standby;
 	int i;
@@ -2693,18 +2689,11 @@ static int find_primary_node(void)
 										  bkinfo->backend_port,
 										  "postgres",
 										  pool_config->sr_check_user,
-										  pool_config->sr_check_password);
+										  pool_config->sr_check_password, true);
 		if (!s)
 		{
 			pool_error("find_primary_node: make_persistent_connection failed");
-
-			/*
-			 * It is possible that a node is down even if
-			 * VALID_BACKEND tells it's valid.  This could happen
-			 * before health checking detects the failure.
-			 * Thus we should continue to look for primary node.
-			 */
-			continue;
+			return -1;
 		}
 		con = s->con;
 #ifdef NOT_USED
@@ -2730,31 +2719,25 @@ static int find_primary_node(void)
 			return -1;
 		}
 #endif
-		if(do_query(con, "SELECT pg_is_in_recovery()", &res, PROTO_MAJOR_V3) == POOL_CONTINUE)
+		status = do_query(con, "SELECT pg_is_in_recovery()",
+						  &res, PROTO_MAJOR_V3);
+		if (res->numrows <= 0)
 		{
-			if (res->numrows <= 0)
-			{
-				pool_log("find_primary_node: do_query returns no rows");
-			}
-			if (res->data[0] == NULL)
-			{
-				pool_log("find_primary_node: do_query returns no data");
-			}
-			if (res->nullflags[0] == -1)
-			{
-				pool_log("find_primary_node: do_query returns NULL");
-			}
-			if (res->data[0] && !strcmp(res->data[0], "t"))
-			{
-				is_standby = true;
-			}
+			pool_log("find_primary_node: do_query returns no rows");
 		}
-		else
+		if (res->data[0] == NULL)
 		{
-			pool_log("find_primary_node: do_query failed");
+			pool_log("find_primary_node: do_query returns no data");
 		}
-		if(res)
-			free_select_result(res);
+		if (res->nullflags[0] == -1)
+		{
+			pool_log("find_primary_node: do_query returns NULL");
+		}
+		if (res->data[0] && !strcmp(res->data[0], "t"))
+		{
+			is_standby = true;
+		}   
+		free_select_result(res);
 		discard_persistent_db_connection(s);
 
 		/*
@@ -2810,7 +2793,7 @@ static int find_primary_node_repeatedly(void)
 /*
 * fork a follow child
 */
-pid_t fork_follow_child(int old_master, int new_primary, int old_primary)
+pid_t fork_follow_child(int old_master, int new_master, int old_primary)
 {
 	pid_t pid;
 	int i;
@@ -2826,7 +2809,7 @@ pid_t fork_follow_child(int old_master, int new_primary, int old_primary)
 			bkinfo = pool_get_node_info(i);
 			if (bkinfo->backend_status == CON_DOWN)
 				trigger_failover_command(i, pool_config->follow_master_command,
-										 old_master, new_primary, old_primary);
+										 old_master, new_master, old_primary);
 		}
 		exit(0);
 	}
